import importlib.metadata
import pathlib

import anywidget
import traitlets

try:
    __version__ = importlib.metadata.version("lloom")
except importlib.metadata.PackageNotFoundError:
    __version__ = "unknown"

_DEV = False # switch to False for production

if _DEV:
  # from `npx vite`
  ESM = "http://localhost:5173/src/index.js?anywidget"
  ESM_select = "http://localhost:5173/src/index_select.js?anywidget"
  CSS = ""
  CSS_select = ""
else:
  # from `npx vite build`
  # Path to static from lloom_ai/src/lloom_ai (the python package)
  bundled_assets_dir = pathlib.Path(__file__).parent / "static"
  ESM = (bundled_assets_dir / "index.js").read_text()
  CSS = (bundled_assets_dir / "index.css").read_text()
  ESM_select = (bundled_assets_dir / "index_select.js").read_text()
  CSS_select = (bundled_assets_dir / "index_select.css").read_text()
  

"""
MATRIX WIDGET
Widget instantiated with anywidget that displays the matrix visualization
"""
class MatrixWidget(anywidget.AnyWidget):
    _esm = ESM
    _css = CSS
    name = traitlets.Unicode().tag(sync=True)

    data = traitlets.Unicode().tag(sync=True)  # syncs the widget's `data` property
    data_items = traitlets.Unicode().tag(sync=True)  # syncs the widget's `data_items` property
    data_items_wide = traitlets.Unicode().tag(sync=True)  # syncs the widget's `data_items_wide` property
    metadata = traitlets.Unicode().tag(sync=True)  # syncs the widget's `metadata` property
  
"""
CONCEPT SELECT WIDGET
Widget instantiated with anywidget that displays the concepts for selection
"""
class ConceptSelectWidget(anywidget.AnyWidget):
    _esm = ESM_select
    _css = CSS_select
    name = traitlets.Unicode().tag(sync=True)

    data = traitlets.Unicode().tag(sync=True)  # syncs the widget's `data` property



"""
llm.py
------
This file contains utility functions for processing calls to LLMs.
"""

# IMPORTS ================================
import openai
import google.generativeai as palm

import time
import random
from pathos.multiprocessing import Pool
import hashlib
import numpy as np

import asyncio
from langchain.chat_models import ChatOpenAI
from langchain.prompts.chat import (
    ChatPromptTemplate,
    SystemMessagePromptTemplate,
    HumanMessagePromptTemplate,
)
import tiktoken

# CONSTANTS ================================
SYS_TEMPLATE = "You are a helpful assistant who helps with identifying patterns in text examples."

RATE_LIMITS = {
    # https://platform.openai.com/account/limits
    # (n_requests, wait_time_secs)
    "gpt-3.5-turbo": (300, 10), # = 300*6 = 1800 rpm (max 10k requests per minute for org)
    "gpt-4": (20, 10), # = 20*6 = 120 rpm (max 10k requests per minute for org)
    "gpt-4-turbo-preview": (20, 10), # = 20*6 = 120 rpm, testing purposes
    "palm": (9, 10), # = 9*6 = 54 rpm (max 90 requests per minute)
}

CONTEXT_WINDOW = {
    # https://platform.openai.com/docs/models
    # Total tokens shared between input and output
    "gpt-3.5-turbo": 16385,  # Max 4096 output tokens
    "gpt-4": 8192, 
    "gpt-4-turbo-preview": 128000,  # Max 4096 output tokens
    "palm": 8000,
}

COSTS = {
    # https://openai.com/pricing
    "gpt-3.5-turbo": [0.0005/1000, 0.0015/1000],
    "gpt-4": [0.03/1000, 0.06/1000],
    "gpt-4-turbo-preview": [0.01/1000, 0.03/1000],
}

EMBED_COSTS = {
    "text-embedding-ada-002": (0.00010/1000),
    "text-embedding-3-small": (0.00002/1000),
    "text-embedding-3-large": (0.00013/1000),
}

def get_system_prompt():
    system_message_prompt = SystemMessagePromptTemplate.from_template(SYS_TEMPLATE)
    return system_message_prompt

def get_token_estimate(text, model_name):
    # Fetch the number of tokens used by a prompt
    encoding = tiktoken.encoding_for_model(model_name)
    tokens = encoding.encode(text)
    num_tokens = len(tokens)
    return num_tokens

def get_token_estimate_list(text_list, model_name):
    # Fetch the number of tokens used by a list of prompts
    token_list = [get_token_estimate(text, model_name) for text in text_list]
    return np.sum(token_list)

def truncate_text_tokens(text, model_name, max_tokens):
    # Truncate a prompt to fit within a maximum number of tokens
    encoding = tiktoken.encoding_for_model(model_name)
    tokens = encoding.encode(text)
    n_tokens = len(tokens)
    if max_tokens is not None and n_tokens > max_tokens:
        # Truncate the prompt
        tokens = tokens[:max_tokens]
        n_tokens = max_tokens
    text = encoding.decode(tokens)
    return text, n_tokens

def calc_cost_by_tokens(model_name, in_tokens, out_tokens):
    # Calculate cost with the tokens and model name
    in_cost = in_tokens * COSTS[model_name][0]
    out_cost = out_tokens * COSTS[model_name][1]
    return in_cost, out_cost

# RETRYING + MULTIPROCESSING ================================

""" 
Defines a Python decorator to avoid API rate limits by retrying with exponential backoff.
From: https://github.com/openai/openai-cookbook/blob/main/examples/How_to_handle_rate_limits.ipynb
"""
def retry_with_exponential_backoff(
    func,
    initial_delay: float = 1,
    exponential_base: float = 2,
    jitter: bool = True,
    max_retries: int = 3,
    errors: tuple = (openai.error.RateLimitError,),
):
    """Retry a function with exponential backoff."""

    def wrapper(*args, **kwargs):
        # Initialize variables
        num_retries = 0
        delay = initial_delay

        # Loop until a successful response or max_retries is hit or an exception is raised
        while True:
            try:
                return func(*args, **kwargs)

            # Retry on specified errors
            except errors as e:
                # Increment retries
                num_retries += 1

                # Check if max retries has been reached
                if num_retries > max_retries:
                    raise Exception(
                        f"Maximum number of retries ({max_retries}) exceeded."
                    )

                # Increment the delay
                delay *= exponential_base * (1 + jitter * random.random())

                # Sleep for the delay
                print("DELAY ", delay)
                time.sleep(delay)

            # Raise exceptions for any errors not specified
            except Exception as e:
                print("Exception", e)

    return wrapper

# CUSTOM LLM API WRAPPERS ================================
def get_res_str(res):
    # Fetch the response string from an LLM response JSON
    return res["choices"][0]["message"]["content"]

@retry_with_exponential_backoff
def base_api_wrapper(cur_prompt, model_name="gpt-3.5-turbo"):
    # Wrapper for calling the base OpenAI API
    res = openai.ChatCompletion.create(
        model=model_name,
        messages=[
                {"role": "system", "content": SYS_TEMPLATE},
                {"role": "user", "content": cur_prompt},
            ]
    )
    res_str = get_res_str(res)
    return res_str

@retry_with_exponential_backoff
def base_api_wrapper_palm(cur_prompt):
    # Uses the chat model (chat-bison-001)
    response = palm.chat(messages=cur_prompt)
    res_str = response.last
    return res_str

""" 
Handles multiprocessing of multiple model queries

Inputs: 
    - prompts: list of prompts to execute
    - multiprocessing: bool. Run multiprocessing if True, otherwise run sequentially

Outputs: List of outputs generated by all models
"""
def multi_query(prompts, model_name, multiprocessing=True):
    if multiprocessing: 
        if model_name != "palm":
            # GPT
            pool = Pool(processes=len(prompts))
            results = pool.map(base_api_wrapper, prompts) # unpacks prompt and model name
            pool.close()
        elif model_name == "palm":
            # PaLM
            batch_size = 90
            delay = 60
            # Run only `batch_size` prompts at a time
            for i in range(0, len(prompts), batch_size):
                cur_prompts = prompts[i:i+batch_size]
                pool = Pool(processes=len(cur_prompts))
                results = pool.map(base_api_wrapper_palm, prompts) # unpacks prompt and model name
                pool.close()
                time.sleep(delay)

    else:
        results = []
        for p, model_name in prompts:
            r = base_api_wrapper(p, model_name)
            results.append(r)
    
    return results

# Mock classes for testing ====================================
class LangChainGeneration():
    # Mock class for a generation within a LangChainResult
    def __init__(self, response):
        self.text = response

class LangChainResult:
    # Mock class for a result from a LangChain API call
    def __init__(self, response):
        gen = LangChainGeneration(response)
        self.generations = [[gen]]
        self.llm_output = {
            "token_usage": {
                "prompt_tokens": 0,
                "completion_tokens": 0,
            }
        }

async def get_test_response(example_ids):
    # Returns a dummy response for testing, currently only for the `themes_to_examples` prompt
    # TODO: Adapt to support other prompts
    full_template = """
    {{
        "pattern_results": [
            {results_str}
        ]
    }}
    """
    def get_ex_json(ex_id):
        template = """
            {{
                "example_id": "{example_id}",
                "rationale": "sample rationale",
                "answer": "A",
            }}
        """
        return template.format(example_id=ex_id)
    
    results = [get_ex_json(ex_id) for ex_id in example_ids]
    results_str = ",\n".join(results)
    res = LangChainResult(response=full_template.format(results_str=results_str))
    return res

def get_prompt_hash(p):
    user_message = p[1].content  # Isolate the user message
    hash = hashlib.sha256(user_message.encode()).hexdigest()
    return hash

def truncate_prompt(prompt, model_name, out_token_alloc):
    # Truncate a prompt to fit within a maximum number of tokens
    max_tokens = CONTEXT_WINDOW[model_name] - out_token_alloc
    for i, message in enumerate(prompt):
        trunc_message, n_tokens = truncate_text_tokens(message.content, model_name, max_tokens)
        prompt[i].content = trunc_message
        max_tokens -= n_tokens
    return prompt

# Main function making calls to LLM
async def multi_query_gpt(chat_model, model_name, prompt_template, arg_dict, batch_num=None, wait_time=None, cache=False, debug=False):
    # Run a single query using LangChain OpenAI Chat API

    # System
    system_message_prompt = get_system_prompt()
    # Human
    human_message_prompt = HumanMessagePromptTemplate.from_template(prompt_template)
    chat_prompt = ChatPromptTemplate.from_messages([system_message_prompt, human_message_prompt])
    
    # Execute prompt
    if wait_time is not None:
        if debug:
            print(f"Batch {batch_num}, wait time {wait_time}")
        await asyncio.sleep(wait_time)  # wait asynchronously
    chat_prompt_formatted = chat_prompt.format_prompt(**arg_dict).to_messages()
    chat_prompt_formatted = truncate_prompt(chat_prompt_formatted, model_name, out_token_alloc=1500)
    prompt_hash = get_prompt_hash(chat_prompt_formatted)

    # Run LLM generation
    try: 
        res_full = await chat_model.agenerate([chat_prompt_formatted])
    except Exception as e:
        print("Error", e)
        return None
    
    # TODO: Add back caching
    return res_full

def process_results(results):
    # Extract just the text generation from the LangChain OpenAI Chat results
    # Insert None if the result is None
    res_text = [(res.generations[0][0].text if res else None) for res in results]
    return res_text

async def multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name, temperature=0, batch_num=None, batched=True, debug=False):
    # Multi-query using LangChain OpenAI Chat API
    chat_model = ChatOpenAI(temperature=temperature, model_name=model_name)
    if debug:
        print("model_name", model_name)
    

    if not batched:
        # Non-batched version
        tasks = [multi_query_gpt(chat_model, model_name, prompt_template, args) for args in arg_dicts]
    else:
        # Batched version
        n_requests, wait_time_secs = RATE_LIMITS[model_name]
        tasks = []
        arg_dict_batches = [arg_dicts[i:i + n_requests] for i in range(0, len(arg_dicts), n_requests)]
        for inner_batch_num, cur_arg_dicts in enumerate(arg_dict_batches):
            chat_model = ChatOpenAI(temperature=temperature, model_name=model_name)
            if batch_num is None:
                wait_time = wait_time_secs * inner_batch_num
            else:
                wait_time = wait_time_secs * batch_num
            if debug:
                wait_time = 0 # Debug mode
            cur_tasks = [multi_query_gpt(chat_model, model_name, prompt_template, arg_dict=args, batch_num=batch_num, wait_time=wait_time) for args in cur_arg_dicts]
            tasks.extend(cur_tasks)

    res_full = await asyncio.gather(*tasks)

    res_text = process_results(res_full)
    return res_text, res_full


# Currently unused
# def call_wrapper(sess, prompt_template, arg_dict, model_name=None, verbose=False):
#     # Wrapper for calling an LLM API, whether the OpenAI API (via base_api_wrapper) or other APIs like LangChain
#     res = None
#     if model_name is None:
#         model_name = sess.model
#     if sess.use_base_api:
#         if arg_dict is None:
#             cur_prompt = prompt_template
#         else:
#             cur_prompt = prompt_template.format(**arg_dict)
#         if verbose:
#             print(cur_prompt)
        
#         if model_name != "palm":
#             # GPT
#             resp = base_api_wrapper(cur_prompt, model_name)
#         elif model_name == "palm":
#             # PaLM
#             resp = base_api_wrapper_palm(cur_prompt)
#         res = resp
#     else:
#         raise Exception("Not implemented yet!")
    
#     # Save raw response to session's llm_cache
#     t = time.strftime('%Y-%m-%d-%H-%M-%S', time.localtime())
#     sess.llm_cache[t] = res
#     return res



# Concept induction concept functions
# =================================================

# Imports
import uuid

# CONCEPT class ================================
class Concept:
    def __init__(self, name, prompt, example_ids, active, summary=None):
        concept_id = str(uuid.uuid4())
        self.id = concept_id
        self.name = name
        self.prompt = prompt
        self.example_ids = example_ids
        self.active = active
        self.summary = summary

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "prompt": self.prompt,
            "example_ids": list(self.example_ids),
            "active": self.active,
            "summary": self.summary
        }



# Concept induction session functions
# =================================================

# Imports
import time
import pandas as pd
import ipywidgets as widgets
import random
import nltk
nltk.download('punkt')
from nltk.tokenize import sent_tokenize


# Local imports
if __package__ is None or __package__ == '':
    # uses current directory visibility
    from concept_induction import *
    from concept import Concept
    from llm import get_token_estimate, EMBED_COSTS
else:
    # uses current package visibility
    from .concept_induction import *
    from .concept import Concept
    from .llm import get_token_estimate, EMBED_COSTS

# WORKBENCH class ================================
class lloom:
    def __init__(
        self,
        in_df: pd.DataFrame,
        doc_id_col: str,
        doc_col: str,
        save_path: str = None,
        debug: bool = False,
    ):
        # Settings
        self.model_name = "gpt-3.5-turbo"
        self.embed_model_name = "text-embedding-3-large"
        self.synth_model_name = "gpt-4-turbo-preview"
        self.score_model_name = "gpt-3.5-turbo"
        self.use_base_api = True
        self.debug = debug  # Whether to run in debug mode

        if save_path is None:
            # Automatically set using timestamp
            t = time.strftime('%Y-%m-%d-%H-%M-%S', time.localtime())
            save_path = f"./exports/{t}"
        self.save_path = save_path

        # Input data
        self.in_df = in_df
        self.doc_id_col = doc_id_col
        self.doc_col = doc_col
        self.df_to_score = in_df  # Default to in_df for concept scoring

        # Output data
        self.saved_dfs = {}  # maps from (step_name, time_str) to df
        self.concepts = {}  # maps from concept_id to Concept 
        self.concept_history = {}  # maps from iteration number to concept dictionary
        self.results = {}  # maps from concept_id to its score_df
        self.df_filtered = None  # Current quotes df
        self.df_bullets = None  # Current bullet points df
        self.select_widget = None  # Widget for selecting concepts
        
        # Cost/Time tracking
        self.time = {}  # Stores time required for each step
        self.cost = {}  # Stores cost incurred by each step
        self.tokens = {
            "in_tokens": [],
            "out_tokens": [],
        }
    
    def save(self):
        # Saves current session to file
        select_widget = self.select_widget
        self.select_widget = None  # Remove widget before saving (can't be pickled)

        t = time.strftime('%Y-%m-%d-%H-%M-%S', time.localtime())
        cur_path = f"{self.save_path}__{t}.pkl"
        with open(cur_path, "wb") as f:
            pickle.dump(self, f)
        print(f"Saved session to {cur_path}")

        self.select_widget = select_widget  # Restore widget after saving

    def get_save_key(self, step_name):
        t = time.strftime('%Y-%m-%d-%H-%M-%S', time.localtime())
        k = (step_name, t)  # Key of step name and current time
        return k

    # Estimate cost of generation for the given params
    def estimate_gen_cost(self, params=None):
        if params is None:
            params = self.auto_suggest_parameters()
            print(f"No parameters provided, so using auto-suggested parameters: {params}")
        # Conservative estimates based on empirical data
        # TODO: change to gather estimates from test cases programmatically
        est_quote_tokens = 40  # Tokens for a quote
        est_bullet_tokens = 10  # Tokens for a bullet point
        est_n_clusters = 4  # Estimate of number of clusters
        est_concept_tokens = 40  # Tokens for one generated concept JSON

        # Filter: generate filter_n_quotes for each doc
        est_cost = {}
        
        filter_in_tokens = np.sum([get_token_estimate(filter_prompt + doc, self.model_name) for doc in self.in_df[self.doc_col].tolist()])
        quotes_tokens_per_doc = params["filter_n_quotes"] * est_quote_tokens 
        filter_out_tokens = quotes_tokens_per_doc * len(self.in_df)
        est_cost["distill_filter"] = calc_cost_by_tokens(self.model_name, filter_in_tokens, filter_out_tokens)

        # Summarize: create n_bullets for each doc
        summ_prompt_tokens = get_token_estimate(summarize_prompt, self.model_name)
        summ_in_tokens = np.sum([(summ_prompt_tokens + quotes_tokens_per_doc) for _ in range(len(self.in_df))])
        bullets_tokens_per_doc = params["summ_n_bullets"] * est_bullet_tokens
        summ_out_tokens = bullets_tokens_per_doc * len(self.in_df)
        est_cost["distill_summarize"] = calc_cost_by_tokens(self.model_name, summ_in_tokens, summ_out_tokens)

        # Cluster: embed each bullet point
        cluster_rate = EMBED_COSTS[self.embed_model_name] 
        cluster_tokens = bullets_tokens_per_doc * len(self.in_df) * cluster_rate
        est_cost["cluster"] = (cluster_tokens, 0)

        # Synthesize: create n_concepts for each of the est_n_clusters
        n_bullets_per_cluster = (params["summ_n_bullets"] * len(self.in_df)) / est_n_clusters
        synth_prompt_tokens = get_token_estimate(synthesize_prompt, self.synth_model_name)
        synth_in_tokens = np.sum([(synth_prompt_tokens + (est_bullet_tokens * n_bullets_per_cluster)) for _ in range(est_n_clusters)])
        synth_out_tokens = params["synth_n_concepts"] * est_n_clusters * est_concept_tokens
        est_cost["synthesize"] = calc_cost_by_tokens(self.synth_model_name, synth_in_tokens, synth_out_tokens)
        
        total_cost = np.sum([c[0] + c[1] for c in est_cost.values()])
        print(f"Estimated cost: {np.round(total_cost, 2)}")
        return est_cost

    # Estimate cost of scoring for the given number of concepts
    def estimate_score_cost(self, n_concepts=None, batch_size=5, get_highlights=True):
        if n_concepts is None:
            active_concepts = self.__get_active_concepts()
            n_concepts = len(active_concepts)
        if get_highlights:
            score_prompt = score_highlight_prompt
        else:
            score_prompt = score_no_highlight_prompt
        
        # TODO: change to gather estimates from test cases programmatically
        est_concept_tokens = 20  # Tokens for concept name + prompt
        est_score_json_tokens = 100  # Tokens for score JSON for one document
            
        score_prompt_tokens = get_token_estimate(score_prompt, self.score_model_name)
        n_batches = math.ceil(len(self.in_df) / batch_size)
        all_doc_tokens = np.sum([get_token_estimate(doc, self.score_model_name) for doc in self.df_to_score[self.doc_col].tolist()])  # Tokens to encode all documents
        score_in_tokens = all_doc_tokens + (n_batches * (score_prompt_tokens + est_concept_tokens))
        score_out_tokens = est_score_json_tokens * n_concepts * len(self.in_df)
        est_cost = calc_cost_by_tokens(self.score_model_name, score_in_tokens, score_out_tokens)

        total_cost = np.sum(est_cost)
        print(f"Estimated cost: {np.round(total_cost, 2)}")
        return est_cost

    def auto_suggest_parameters(self, sample_size=None, target_n_concepts=20, debug=False):
        # Suggests concept generation parameters based on rough heuristics
        # TODO: Use more sophisticated methods to suggest parameters
        if sample_size is not None:
            sample_docs = self.in_df[self.doc_col].sample(sample_size).tolist()
        else:
            sample_docs = self.in_df[self.doc_col].tolist()

        # Get number of sentences in each document
        n_sents = [len(sent_tokenize(doc)) for doc in sample_docs]
        avg_n_sents = int(np.median(n_sents))
        if debug:
            print(f"N sentences: Median={avg_n_sents}, Std={np.std(n_sents):0.2f}")
        quote_per_sent = 0.75  # Average number of quotes per sentence
        filter_n_quotes = max(1, math.ceil(avg_n_sents * quote_per_sent))

        bullet_per_quote = 0.75  # Average number of bullet points per quote
        summ_n_bullets = max(1, math.floor(filter_n_quotes * bullet_per_quote))

        est_n_clusters = 3
        synth_n_concepts = math.floor(target_n_concepts / est_n_clusters)
        params = {
            "filter_n_quotes": filter_n_quotes,
            "summ_n_bullets": summ_n_bullets,
            "synth_n_concepts": synth_n_concepts,
        }
        return params
    
    def summary(self, show_detail=True):
        # Time
        total_time = np.sum(list(self.time.values()))
        print(f"Total time: {total_time:0.2f} sec ({(total_time/60):0.2f} min)")
        if show_detail:
            for step_name, time in self.time.items():
                print(f"\t{step_name}: {time:0.2f} sec")

        # Cost
        total_cost = np.sum(list(self.cost.values()))
        print(f"\n\nTotal cost: {total_cost:0.2f}")
        if show_detail:
            for step_name, cost in self.cost.items():
                print(f"\t{step_name}: {cost:0.2f}")

        # Tokens
        in_tokens = np.sum(self.tokens["in_tokens"])
        out_tokens = np.sum(self.tokens["out_tokens"])
        total_tokens =  in_tokens + out_tokens
        print(f"\n\nTokens: total={total_tokens}, in={in_tokens}, out={out_tokens}")

    def show_selected(self):
        active_concepts = self.__get_active_concepts()
        print(f"Active concepts (n={len(active_concepts)}):")
        for c_id, c in active_concepts.items():
            print(f"- {c.name}: {c.prompt}")

    # HELPER FUNCTIONS ================================
    async def gen(self, seed=None, params=None, n_synth=1, debug=True):
        if params is None:
            params = self.auto_suggest_parameters(debug=debug)
            if debug:
                print(f"Auto-suggested parameters: {params}")
                self.estimate_gen_cost(params)

        # Run concept generation
        filter_n_quotes = params["filter_n_quotes"]
        if filter_n_quotes > 1:
            df_filtered = await distill_filter(
                text_df=self.in_df, 
                doc_col=self.doc_col,
                doc_id_col=self.doc_id_col,
                model_name=self.model_name,
                n_quotes=params["filter_n_quotes"],
                seed=seed,
                sess=self,
            )
            self.df_to_score = df_filtered  # Change to use filtered df for concept scoring
            self.df_filtered = df_filtered
            if debug:
                print("df_filtered")
                display(df_filtered)
        else:
            # Just use original df to generate bullets
            self.df_filtered = self.in_df
        
        df_bullets = await distill_summarize(
            text_df=df_filtered, 
            doc_col=self.doc_col,
            doc_id_col=self.doc_id_col,
            model_name=self.model_name,
            n_bullets=params["summ_n_bullets"],
            seed=seed,
            sess=self,
        )
        self.df_bullets = df_bullets
        if debug:
            print("df_bullets")
            display(df_bullets)
        
        df_cluster_in = df_bullets
        synth_doc_col = self.doc_col
        synth_n_concepts = params["synth_n_concepts"]
        concept_col_prefix = "concept"
        # Perform synthesize step n_synth times
        for i in range(n_synth):
            self.concepts = {}
            df_cluster = await cluster(
                text_df=df_cluster_in, 
                doc_col=synth_doc_col,
                doc_id_col=self.doc_id_col,
                embed_model_name=self.embed_model_name,
                sess=self,
            )
            if debug:
                print("df_cluster")
                display(df_cluster)
            
            df_concepts = await synthesize(
                cluster_df=df_cluster, 
                doc_col=synth_doc_col,
                doc_id_col=self.doc_id_col,
                model_name=self.synth_model_name,
                concept_col_prefix=concept_col_prefix,
                n_concepts=synth_n_concepts,
                pattern_phrase="unique topic",
                seed=seed,
                sess=self,
            )

            self.concept_history[i] = self.concepts
            if debug:
                # Print results
                print(f"synthesize {i + 1}: (n={len(self.concepts)})")
                for k, c in self.concepts.items():
                    print(f'- Concept {k}:\n\t{c.name}\n\t- Prompt: {c.prompt}')
            
            # Update synthesize params for next iteration
            df_concepts["synth_doc_col"] = df_concepts[f"{concept_col_prefix}_name"] + ": " + df_concepts[f"{concept_col_prefix}_prompt"]
            df_cluster_in = df_concepts
            synth_doc_col = "synth_doc_col"
            synth_n_concepts = math.floor(synth_n_concepts * 0.75)

    def __concepts_to_json(self):
        concept_dict = {c_id: c.to_dict() for c_id, c in self.concepts.items()}
        # Get examples from example IDs
        for c_id, c in concept_dict.items():
            ex_ids = c["example_ids"]
            in_df = self.df_filtered.copy()
            in_df[self.doc_id_col] = in_df[self.doc_id_col].astype(str)
            examples = in_df[in_df[self.doc_id_col].isin(ex_ids)][self.doc_col].tolist()
            c["examples"] = examples
        return json.dumps(concept_dict)
    
    def select(self):
        concepts_json = self.__concepts_to_json()
        w = get_select_widget(concepts_json)
        self.select_widget = w
        return w

    def __get_active_concepts(self):
        # Update based on widget
        if self.select_widget is not None:
            widget_data = json.loads(self.select_widget.data)
            for c_id, c in self.concepts.items():
                widget_active = widget_data[c_id]["active"]
                c.active = widget_active
        return {c_id: c for c_id, c in self.concepts.items() if c.active}

    # Score the specified concepts
    # Only score the concepts that are active
    async def score(self, c_ids=None, get_highlights=True, ignore_existing=True):
        concepts = {}
        active_concepts = self.__get_active_concepts()
        if c_ids is None:
            # Score all active concepts
            for c_id, c in active_concepts.items():
                concepts[c_id] = c
        else:
            # Score only the specified concepts
            for c_id in c_ids:
                if c_id in active_concepts:
                    concepts[c_id] = active_concepts[c_id]
        
        # Ignore concepts that already have existing results
        if ignore_existing:
            concepts = {c_id: c for c_id, c in concepts.items() if c_id not in self.results}

        # Run usual scoring; results are stored to self.results within the function
        score_df = await score_concepts(
            text_df=self.df_to_score, 
            text_col=self.doc_col, 
            doc_id_col=self.doc_id_col,
            concepts=concepts,
            model_name=self.score_model_name,
            get_highlights=get_highlights,
            sess=self,
        )

        return score_df

    def __get_concept_from_name(self, name):
        if name == "Outlier":
            return Concept(name="Outlier", prompt=OUTLIER_CRITERIA, example_ids=[], active=True)
        for c_id, c in self.concepts.items():
            if c.name == name:
                return c
        return None
    
    def get_score_df(self):
        active_concepts = self.__get_active_concepts()
        active_score_dfs = [self.results[c_id] for c_id in active_concepts.keys() if c_id in self.results]
        score_df = pd.concat(active_score_dfs)
        score_df = score_df.rename(columns={"doc_id": self.doc_id_col})
        return score_df

    def __get_concept_highlights(self, c, threshold=0.75, highlight_col="highlight", lim=3):
        if c.name == "Outlier":
            return []
        if c.id not in self.results:
            return []
        score_df = self.results[c.id].copy()
        score_df = score_df[score_df["score"] > threshold]
        highlights = score_df[highlight_col].tolist()
        # shuffle highlights
        random.shuffle(highlights)
        if lim is not None:
            highlights = highlights[:lim]
        return highlights

    def __get_rep_examples(self, c):
        if c.name == "Outlier":
            return []
        if c.id not in self.results:
            return []
        df = self.df_filtered.copy()
        df[self.doc_id_col] = df[self.doc_id_col].astype(str)
        ex_ids = c.example_ids
        ex = df[df[self.doc_id_col].isin(ex_ids)][self.doc_col].tolist()
        return ex

    def __get_df_for_export(self, item_df, threshold=0.75, include_outliers=False):
        # Prepares a dataframe meant for exporting the current session results
        # Includes concept, criteria, summary, representative examples, prevalence, and highlights
        matched = item_df[(item_df.concept_score_orig > threshold)]
        if not include_outliers:
            matched = matched[item_df.concept != "Outlier"]

        df = matched.groupby(by=["id", "concept"]).count().reset_index()[["concept", self.doc_col]]
        concepts = [self.__get_concept_from_name(c_name) for c_name in df.concept.tolist()]
        df["criteria"] = [c.prompt for c in concepts]
        df["summary"] = [c.summary for c in concepts]
        df["rep_examples"] = [self.__get_rep_examples(c) for c in concepts]
        df["highlights"] = [self.__get_concept_highlights(c, threshold) for c in concepts]
        df = df.rename(columns={self.doc_col: "n_matches"})
        df["prevalence"] = np.round(df["n_matches"] / len(self.in_df), 2)
        df = df[["concept", "criteria", "summary", "rep_examples", "prevalence", "highlights"]]
        return df
        

    def vis(self, cols_to_show=[], custom_groups={}, show_highlights=True, norm_by="concept", export_df=False):
        active_concepts = self.__get_active_concepts()
        score_df = self.get_score_df()

        widget, matrix_df, item_df, item_df_wide = visualize(
            in_df=self.in_df,
            score_df=score_df,
            doc_col=self.doc_col,
            doc_id_col=self.doc_id_col,
            score_col="score",
            df_filtered=self.df_filtered,
            df_bullets=self.df_bullets,
            concepts=active_concepts,
            cols_to_show=cols_to_show,
            custom_groups=custom_groups,
            show_highlights=show_highlights,
            norm_by=norm_by,
        )
        if export_df:
            return self.__get_df_for_export(item_df)
        
        return widget

    def export_df(self):
        return self.vis(export_df=True)

    async def add(self, name, prompt, ex_ids=[], get_highlights=True):
        # Add concept
        c = Concept(name=name, prompt=prompt, example_ids=ex_ids, active=True)
        self.concepts[c.id] = c

        # Update widget
        self.select_widget = self.select()

        # Run scoring
        cur_score_df = await self.score(c_ids=[c.id], get_highlights=get_highlights)
        
        # Store results
        self.results[c.id] = cur_score_df
    
    
    async def edit(self):
        raise NotImplementedError("Edit function not yet implemented")



# Main concept induction library functions
# =================================================

# Imports
import yaml
import pandas as pd
import time
from tqdm.asyncio import tqdm_asyncio
import numpy as np
import math
import json
import uuid
import sys
import textwrap
from itertools import chain
import pickle
import ipywidgets as widgets

# Clustering
from bertopic import BERTopic
from bertopic.backend import OpenAIBackend
from hdbscan import HDBSCAN

# Local imports
if __package__ is None or __package__ == '':
    # uses current directory visibility
    from llm import multi_query_gpt_wrapper, calc_cost_by_tokens
    from prompts import *
    from concept import Concept
    from __init__ import MatrixWidget, ConceptSelectWidget
else:
    # uses current package visibility
    from .llm import multi_query_gpt_wrapper, calc_cost_by_tokens
    from .prompts import *
    from .concept import Concept
    from .__init__ import MatrixWidget, ConceptSelectWidget

# CONSTANTS ================================
NAN_SCORE = -0.01  # Numerical score to use in place of NaN values for matrix viz
OUTLIER_CRITERIA = "Did the example not match any of the above concepts?"


# HELPER functions ================================

def json_load(s, top_level_key=None):
    # Attempts to safely load a JSON from a string response from the LLM
    if s is None:
        return None
    json_start = s.find("{")
    json_end = s.rfind("}")
    s_trimmed = s[json_start:(json_end + 1)]
    
    try:
        cur_dict = yaml.safe_load(s_trimmed)
        
        if (top_level_key is not None) and top_level_key in cur_dict:
            cur_dict = cur_dict[top_level_key]
            return cur_dict
        return cur_dict
    except:
        print(f"ERROR json_load on: {s}")
        return None

def pretty_print_dict(d):
    # Print all elements within a provided dictionary
    return "\n\t".join([f"{k}: {v}" for k, v in d.items()])

def pretty_print_dict_list(d_list):
    # Print all dictionaries in a list of dictionaries
    return "\n\t" + "\n\t".join([pretty_print_dict(d) for d in d_list])

def cluster_helper(in_df, doc_col, doc_id_col, min_cluster_size, cluster_id_col, embed_model_name):
    # OpenAI embeddings with HDBSCAN clustering
    embedding_model = OpenAIBackend(embed_model_name)

    hdbscan_model = HDBSCAN(
        min_cluster_size=min_cluster_size, 
        metric='euclidean', 
        cluster_selection_method='leaf', 
        prediction_data=True
    )
    topic_model = BERTopic(
        embedding_model=embedding_model, 
        hdbscan_model=hdbscan_model
    )

    id_vals = in_df[doc_id_col].tolist()
    text_vals = in_df[doc_col].tolist()
    clusters, probs = topic_model.fit_transform(text_vals)
    rows = list(zip(id_vals, text_vals, clusters)) # id_col, text_col, cluster_id_col
    cluster_df = pd.DataFrame(rows, columns=[doc_id_col, doc_col, cluster_id_col])
    cluster_df = cluster_df.sort_values(by=[cluster_id_col])
    
    return cluster_df

def save_progress(sess, df, step_name, start, res, model_name, debug=True):
    # Save df to session
    if (sess is not None) and (df is not None):
        k = sess.get_save_key(step_name)
        sess.saved_dfs[k] = df

    # Gets timing
    get_timing(start, step_name, sess, debug=debug)

    # Gets cost
    calc_cost(res, model_name, step_name, sess, debug=debug)

def get_timing(start, step_name, sess, debug=True):
    if start is None:
        return
    elapsed = time.time() - start
    if debug:
        print(f"Total time: {elapsed:0.2f} sec")
    if sess is not None:
        # Save to session if provided
        k = sess.get_save_key(step_name)
        sess.time[k] = elapsed

def calc_cost(results, model_name, step_name, sess, debug=True):
    # Calculate cost with API results and model name
    if results is None:
        return
    # Cost estimation
    in_tokens = np.sum([res.llm_output["token_usage"]["prompt_tokens"] for res in results])
    out_tokens = np.sum([res.llm_output["token_usage"]["completion_tokens"] for res in results])
    in_token_cost, out_token_cost = calc_cost_by_tokens(model_name, in_tokens, out_tokens)
    total_cost = in_token_cost + out_token_cost
    if debug:
        print(f"\nTotal: {total_cost} | In: {in_token_cost} | Out: {out_token_cost}")
    if sess is not None:
        # Save to session if provided
        sess.tokens["in_tokens"].append(in_tokens)
        sess.tokens["out_tokens"].append(out_tokens)
        k = sess.get_save_key(step_name)
        sess.cost[k] = total_cost

def filter_empty_rows(df, text_col_name):
    # Remove rows where the specified column is empty
    df_out = df[df[text_col_name].apply(lambda x: len(x) > 0)]
    return df_out

# CORE functions ================================

# Input: 
# - text_df: DataFrame (columns: doc_id, doc)
# Parameters: model_name, n_quotes, seed
# Output: quote_df (columns: doc_id, quote)
async def distill_filter(text_df, doc_col, doc_id_col, model_name, n_quotes=3, seed=None, sess=None):
    # Filtering operates on provided text
    start = time.time()

    # Filter to non-empty rows
    text_df = filter_empty_rows(text_df, doc_col)

    # Prepare prompts
    filtered_ex = []
    rows = []
    if seed is not None:
        seeding_phrase = f"related to {seed}"
    else:
        seeding_phrase = "most important"
    arg_dicts = [
        {
            "ex": ex, 
            "n_quotes": n_quotes, 
            "seeding_phrase": seeding_phrase.upper()
        } for ex in text_df[doc_col].tolist()
    ]
    
    # Run prompts
    prompt_template = filter_prompt
    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)

    # Process results
    ex_ids = [ex_id for ex_id in text_df[doc_id_col].tolist()]
    for ex_id, res in zip(ex_ids, res_text):
        cur_filtered_list = json_load(res, top_level_key="relevant_quotes")
        if cur_filtered_list is not None:
            cur_filtered = "\n".join(cur_filtered_list)
            filtered_ex.append(cur_filtered)
            rows.append([ex_id, cur_filtered])
    quote_df = pd.DataFrame(rows, columns=[doc_id_col, doc_col])
    
    save_progress(sess, quote_df, step_name="distill_filter", start=start, res=res_full, model_name=model_name)
    return quote_df


# Input: text_df (columns: doc_id, doc) 
#   --> text could be original or filtered (quotes)
# Parameters: n_bullets, n_words_per_bullet, seed
# Output: bullet_df (columns: doc_id, bullet)
async def distill_summarize(text_df, doc_col, doc_id_col, model_name, n_bullets="2-4", n_words_per_bullet="5-8", seed=None, sess=None):
    # Summarization operates on text_col
    start = time.time()

    # Filter to non-empty rows
    text_df = filter_empty_rows(text_df, doc_col)

    # Prepare prompts
    rows = []
    arg_dicts = []
    all_ex_ids = []

    if seed is not None:
        seeding_phrase = f"related to {seed}"
    else:
        seeding_phrase = ""
    for _, row in text_df.iterrows():
        ex = row[doc_col]
        ex_id = row[doc_id_col]
        if len(ex) == 0:
            # Handle if filtered example is empty
            rows.append([ex_id, ""])
            continue
        arg_dict = {
            "ex": ex, 
            "seeding_phrase": seeding_phrase.upper(), 
            "n_bullets": n_bullets, 
            "n_words": n_words_per_bullet
        }
        arg_dicts.append(arg_dict)
        all_ex_ids.append(ex_id)
    
    # Run prompts
    prompt_template = summarize_prompt
    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)

    # Process results
    for ex_id, res in zip(all_ex_ids, res_text):
        cur_bullets_list = json_load(res, top_level_key="bullets")
        if cur_bullets_list is not None:
            for bullet in cur_bullets_list:
                # Expand each bullet into its own row
                rows.append([ex_id, bullet])
    bullet_df = pd.DataFrame(rows, columns=[doc_id_col, doc_col])

    save_progress(sess, bullet_df, step_name="distill_summarize", start=start, res=res_full, model_name=model_name)
    return bullet_df


# Input: text_df (columns: doc_id, doc) 
#   --> text could be original, filtered (quotes), and/or summarized (bullets)
# Parameters: n_clusters
# Output: cluster_df (columns: doc_id, doc, cluster_id)
async def cluster(text_df, doc_col, doc_id_col, cluster_id_col="cluster_id", min_cluster_size=None, embed_model_name="text-embedding-ada-002", batch_size=20, randomize=False, sess=None):
    # Clustering operates on text_col
    start = time.time()

    # Filter to non-empty rows
    text_df = filter_empty_rows(text_df, doc_col)

    # Auto-set parameters
    n_items = len(text_df)
    if min_cluster_size is None:
        min_cluster_size = max(3, int(n_items/10))
    n_batches = math.ceil(n_items / batch_size)

    # Generate clusters
    if randomize:
        # Randomize the bulletpoints
        text_df = text_df.sample(frac=1)  # shuffle order
        cluster_df = text_df.copy()
        cluster_ids = [
            [i for _ in range(batch_size)] 
            for i in range(n_batches)
        ]
        cluster_ids = list(chain.from_iterable(cluster_ids))[:n_items]
        cluster_df[cluster_id_col] = cluster_ids
    else:
        # Cluster and group by clusters
        cluster_df = cluster_helper(text_df, doc_col, doc_id_col, min_cluster_size=min_cluster_size, cluster_id_col=cluster_id_col, embed_model_name=embed_model_name)

    save_progress(sess, cluster_df, step_name="cluster", start=start, res=None, model_name=None)
    return cluster_df


def dict_to_json(examples):
    # Internal helper to convert examples to json for prompt
    examples_json = json.dumps(examples)
    # Escape curly braces to avoid the system interpreting as template formatting
    examples_json = examples_json.replace("{", "{{")
    examples_json = examples_json.replace("}", "}}")
    return examples_json

# Input: cluster_df (columns: doc_id, doc, cluster_id)
# Parameters: n_concepts
# Output: 
# - concepts: dict (concept_id -> concept dict)
# - concept_df: DataFrame (columns: doc_id, doc, concept_id, concept_name, concept_prompt)
async def synthesize(cluster_df, doc_col, doc_id_col, model_name, cluster_id_col="cluster_id", concept_col_prefix="concept", n_concepts=None, batch_size=None, verbose=True, pattern_phrase="unifying pattern", dedupe=True, seed=None, sess=None):
    # Synthesis operates on "doc" column for each cluster_id
    # Concept object is created for each concept
    start = time.time()
    
    # Filter to non-empty rows
    cluster_df = filter_empty_rows(cluster_df, doc_col)

    # Auto-set parameters
    def get_n_concepts_phrase(cur_set):
        if n_concepts is None:
            cur_n_concepts = math.ceil(len(cur_set)/3)
        else:
            cur_n_concepts = n_concepts
        if cur_n_concepts > 1:
            return f"up to {cur_n_concepts} {pattern_phrase}s"
        else:
            return f"{cur_n_concepts} {pattern_phrase}"
    
    # Prepare prompts
    # Create prompt arg dictionary with example IDs
    if seed is not None:
        seed_phrase = f"If possible, please make the patterns RELATED TO {seed.upper()}."
    else:
        seed_phrase = ""
    arg_dicts = []
    cluster_ids = cluster_df[cluster_id_col].unique()
    cluster_dfs = {}  # Store each cluster's dataframe by cluster_id
    ex_id_to_ex = {(str(row[doc_id_col]), row[cluster_id_col]): row[doc_col] for _, row in cluster_df.iterrows()}  # Map example IDs to example text
    for cluster_id in cluster_ids:
        # Iterate over cluster IDs to get example sets
        cur_df = cluster_df[cluster_df[cluster_id_col] == cluster_id]
        cluster_dfs[cluster_id] = cur_df
        if batch_size is not None:
            # Split into batches
            n_batches = math.ceil(len(cur_df) / batch_size)
            for i in range(n_batches):
                cur_batch_df = cur_df.iloc[i*batch_size:(i+1)*batch_size]
                ex_dicts = [{"example_id": row[doc_id_col], "example": row[doc_col]} for _, row in cur_batch_df.iterrows()]
                ex_dicts_json = dict_to_json(ex_dicts)
                arg_dict = {
                    "examples": ex_dicts_json,
                    "n_concepts_phrase": get_n_concepts_phrase(cur_df),
                    "seed_phrase": seed_phrase
                }
                arg_dicts.append(arg_dict)
        else:
            # Handle unbatched case
            ex_dicts = [{"example_id": row[doc_id_col], "example": row[doc_col]} for _, row in cur_df.iterrows()]
            ex_dicts_json = dict_to_json(ex_dicts)
            arg_dict = {
                "examples": ex_dicts_json,
                "n_concepts_phrase": get_n_concepts_phrase(cur_df),
                "seed_phrase": seed_phrase
            }
            arg_dicts.append(arg_dict)

    # Run prompts
    prompt_template = synthesize_prompt
    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)

    # Process results
    concepts = {}
    rows = []
    for cur_cluster_id, res in zip(cluster_ids, res_text):
        cur_concepts = json_load(res, top_level_key="patterns")
        if cur_concepts is not None:
            # concepts.extend(cur_concepts)
            # cur_examples_dict = {ex_to_id[cur_ex]: cur_ex for cur_ex in cur_examples}
            
            for concept_dict in cur_concepts:
                ex_ids = concept_dict["example_ids"]
                ex_ids = set(ex_ids) # remove duplicates
                concept = Concept(
                    name=concept_dict["name"],
                    prompt=concept_dict["prompt"],
                    example_ids=ex_ids,
                    active=False,
                )
                concepts[concept.id] = concept
                
                for ex_id in ex_ids:
                    # doc_id, text, concept_id, concept_name, concept_prompt
                    cur_key = (ex_id, cur_cluster_id)
                    if cur_key in ex_id_to_ex:
                        row = [ex_id, ex_id_to_ex[cur_key], concept.id, concept.name, concept.prompt]
                        rows.append(row)
            if verbose:
                examples = cluster_dfs[cur_cluster_id][doc_col].tolist()
                concepts_formatted = pretty_print_dict_list(cur_concepts)
                print(f"\n\nInput examples: {examples}\nOutput concepts: {concepts_formatted}")
    # doc_id, text, concept_id, concept_name, concept_prompt
    concept_df = pd.DataFrame(rows, columns=[doc_id_col, doc_col, concept_col_prefix, f"{concept_col_prefix}_name", f"{concept_col_prefix}_prompt"])

    concept_df[f"{concept_col_prefix}_namePrompt"] = concept_df[f"{concept_col_prefix}_name"] + ": " + concept_df[f"{concept_col_prefix}_prompt"]
    if dedupe:
        concept_df = dedupe_concepts(concept_df, concept_col=f"{concept_col_prefix}_namePrompt")

    save_progress(sess, concept_df, step_name="synthesize", start=start, res=res_full, model_name=model_name)
    # Save to session if provided
    if sess is not None:
        for c_id, c in concepts.items():
            sess.concepts[c_id] = c
    return concept_df

def dedupe_concepts(df, concept_col):
    # Remove duplicate concept rows
    return df.drop_duplicates(subset=[concept_col])

def pretty_print_merge_results(merged):
    for m in merged:
        orig_concepts = m["original_themes"]
        print(f"[{orig_concepts}] --> {m['merged_theme_name']}: {m['merged_theme_prompt']}")

# Input: concept_df (columns: doc_id, text, concept_id, concept_name, concept_prompt)
# Parameters: n_concepts
# Output: 
# - concepts: dict (concept_id -> concept dict)
# - concept_df: DataFrame (columns: doc_id, text, concept_id, concept_name, concept_prompt)
async def review(concepts, concept_df, concept_col, concept_col_prefix, model_name, debug=True, sess=None):
    # Model is asked to review the provided set of concepts
    concepts_out, concept_df_out, removed = await review_remove(concepts, concept_df, concept_col, concept_col_prefix, model_name=model_name, sess=sess)
    concepts_out, concept_df_out, merged = await review_merge(concepts_out, concept_df_out, concept_col, concept_col_prefix, model_name=model_name, sess=sess)

    if debug:
        print(f"Removed ({len(removed)}):\n{removed}")
        print(f"Merged ({len(merged)}):")
        pretty_print_merge_results(merged)

    # TODO: ask model to filter to the "best" N concepts
    return concepts_out, concept_df_out


# Model removes concepts that are too specific or too general
# Input: concept_df (columns: doc_id, text, concept_id, concept_name, concept_prompt)
# Parameters: n_concepts
# Output: 
# - concepts: dict (concept_id -> concept dict)
# - concept_df: DataFrame (columns: doc_id, text, concept_id, concept_name, concept_prompt)
async def review_remove(concepts, concept_df, concept_col, concept_col_prefix, model_name, sess):
    concepts = concepts.copy()  # Make a copy of the concepts dict to avoid modifying the original
    start = time.time()
    concept_name_col = f"{concept_col_prefix}_name"

    concepts_list = concept_df[concept_col].tolist()
    concepts_list = [f"- {c}" for c in concepts_list]
    concepts_list_str = "\n".join(concepts_list)
    arg_dicts = [{
        "themes": concepts_list_str,
    }]

    # Run prompts
    prompt_template = review_remove_prompt
    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)

    # Process results
    res = res_text[0]
    concepts_to_remove = json_load(res, top_level_key="remove")

    concept_df_out = concept_df.copy()
    concept_df_out = concept_df_out[~concept_df_out[concept_name_col].isin(concepts_to_remove)]
    c_ids_to_remove = []
    for c_id, c in concepts.items():
        if c['name'] in concepts_to_remove:
            c_ids_to_remove.append(c_id)
    for c_id in c_ids_to_remove:
        concepts.pop(c_id, None)

    save_progress(sess, concept_df_out, step_name="review_remove", start=start, res=res_full, model_name=model_name)
    return concepts, concept_df_out, concepts_to_remove

def get_concept_by_name(concepts, concept_name):
    for c_id, c in concepts.items():
        if c['name'] == concept_name:
            return c_id, c
    return None, None

# Model merges concepts that are similar or overlapping
# Input: concept_df (columns: doc_id, text, concept_id, concept_name, concept_prompt)
# Parameters: n_concepts
# Output: 
# - concepts: dict (concept_id -> concept dict)
# - concept_df: DataFrame (columns: doc_id, text, concept_id, concept_name, concept_prompt)
async def review_merge(concepts, concept_df, concept_col, concept_col_prefix, model_name, sess):
    concepts = concepts.copy()  # Make a copy of the concepts dict to avoid modifying the original
    start = time.time()
    concept_name_col = f"{concept_col_prefix}_name"
    concept_prompt_col = f"{concept_col_prefix}_prompt"

    concepts_list = concept_df[concept_col].tolist()
    concepts_list = [f"- {c}" for c in concepts_list]
    concepts_list_str = "\n".join(concepts_list)
    arg_dicts = [{
        "themes": concepts_list_str,
    }]

    # Run prompts
    prompt_template = review_merge_prompt
    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)

    # Process results
    res = res_text[0]
    concepts_to_merge = json_load(res, top_level_key="merge")

    concept_df_out = concept_df.copy()

    # Remove all original concepts
    # Add new merged concepts
    concepts_to_remove = []
    c_ids_to_remove = []
    for merge_result in concepts_to_merge:
        orig_concepts = merge_result["original_themes"]
        # Only allow merging pairs
        if len(orig_concepts) != 2:
            continue
        
        # Don't allow duplicates (in prior merge pairs)
        for orig_concept in orig_concepts:
            if orig_concept in concepts_to_remove:
                continue

        concepts_to_remove.extend(orig_concepts)

        # Get original concept IDs and example IDs
        merged_example_ids = []
        for orig_concept in orig_concepts:
            c_id, c = get_concept_by_name(concepts, orig_concept)
            if c is not None:
                c_ids_to_remove.append(c_id)
                merged_example_ids.extend(c["example_ids"])
        
        # Create new merged concept in dict
        new_concept_name = merge_result["merged_theme_name"]
        new_concept_prompt = merge_result["merged_theme_prompt"]
        new_concept_id = str(uuid.uuid4())
        concepts[new_concept_id] = {
            "name": new_concept_name,
            "prompt": new_concept_prompt,
            "example_ids": merged_example_ids,
        }

        # Replace prior df row with new merged concept
        for orig_concept in orig_concepts:
            concept_df_out.loc[concept_df_out[concept_name_col]==orig_concept, concept_name_col] = new_concept_name
            concept_df_out.loc[concept_df_out[concept_name_col]==orig_concept, concept_prompt_col] = new_concept_prompt
            concept_df_out.loc[concept_df_out[concept_name_col]==orig_concept, concept_col] = f"{new_concept_name}: {new_concept_prompt}"
        
    for c_id in c_ids_to_remove:
        concepts.pop(c_id, None)

    save_progress(sess, concept_df_out, step_name="review_merge", start=start, res=res_full, model_name=model_name)
    return concepts, concept_df_out, concepts_to_merge


def get_ex_batch_args(df, text_col, doc_id_col, concept_name, concept_prompt):
    ex = get_examples_dict(df, text_col, doc_id_col)
    examples_json = dict_to_json(ex)
    arg_dict = {
        "examples_json": examples_json,
        "pattern_name": concept_name,
        "pattern_prompt": concept_prompt,
        "example_ids": list(df[doc_id_col])
    }
    return arg_dict

def get_examples_dict(cur_df, text_col, doc_id_col):
    # Helper to get examples from cur_df in dictionary form for JSON in prompt
    ex_list = []
    for i, row in cur_df.iterrows():
        ex_dict = {
            "example_id": row[doc_id_col],
            "example_text": row[text_col],
        }
        ex_list.append(ex_dict)

    ex = {"cur_examples": ex_list}
    return ex

def parse_bucketed_score(x):
    # Internal helper to parse bucketed score from LLM response to numerical result
    answer_scores = {
        "A": 1.0,
        "B": 0.75,
        "C": 0.5,
        "D": 0.25,
        "E": 0.0,
    }
    if len(x) > 1:
        x = x[0]
    if x not in answer_scores.keys():
        return NAN_SCORE
    return answer_scores[x]

def get_score_df(res, in_df, concept, concept_id, text_col, doc_id_col, get_highlights):
    # Cols: doc_id, text, concept_id, concept_name, concept_prompt, score, highlight
    res_dict = json_load(res, top_level_key="pattern_results")
    OUT_COLS = ["doc_id", "text", "concept_id", "concept_name", "concept_prompt", "score", "rationale", "highlight"]
    concept_name = concept.name
    concept_prompt = concept.prompt
    if res_dict is not None:
        rows = []
        for ex in res_dict:
            if "answer" in ex:
                ans = parse_bucketed_score(ex["answer"])
            else:
                ans = NAN_SCORE
            
            if "example_id" in ex:
                doc_id = ex["example_id"]
                text_list = in_df[in_df[doc_id_col] == doc_id][text_col].tolist()
                if len(text_list) > 0:
                    # Document is in the dataset
                    text = text_list[0]

                    if "rationale" in ex:
                        rationale = ex["rationale"]
                    else:
                        rationale = ""   # Set rationale to empty string

                    if get_highlights and ("quote" in ex):
                        row = [doc_id, text, concept_id, concept_name, concept_prompt, ans, rationale, ex["quote"]]
                    else:
                        row = [doc_id, text, concept_id, concept_name, concept_prompt, ans, rationale, ""]  # Set quote to empty string
                    rows.append(row)
        
        out_df = pd.DataFrame(rows, columns=OUT_COLS)
        return out_df
    else:
        out_df = in_df.copy()
        out_df["doc_id"] = out_df[doc_id_col]
        out_df["text"] = out_df[text_col]
        out_df["concept_id"] = concept_id
        out_df["concept_name"] = concept_name
        out_df["concept_prompt"] = concept_prompt
        out_df["score"] = NAN_SCORE
        out_df["rationale"] = ""
        out_df["highlight"] = ""
        return out_df[OUT_COLS]

# Performs scoring for one concept
async def score_helper(concept, batch_i, concept_id, df, text_col, doc_id_col, model_name, batch_size, get_highlights, sess):
    # TODO: add support for only a sample of examples
    # TODO: set consistent concept IDs for reference

    # Prepare batches of input arguments
    indices = range(0, len(df), batch_size)
    ex_ids = [str(x) for x in df[doc_id_col].tolist()]
    ex_id_sets = [ex_ids[i:i+batch_size] for i in indices]
    in_dfs = [df[df[doc_id_col].isin(cur_ex_ids)] for cur_ex_ids in ex_id_sets]
    arg_dicts = [
        get_ex_batch_args(df, text_col, doc_id_col, concept.name, concept.prompt) for df in in_dfs
    ]

    # Run prompts in parallel to score each example
    if get_highlights:
        prompt_template = score_highlight_prompt
    else:
        prompt_template = score_no_highlight_prompt
    results, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name, batch_num=batch_i)

    # Parse results
    # Cols: doc_id, text, concept_id, concept_name, concept_prompt, score, highlight
    score_df = None
    for i, res in enumerate(results):
        in_df = in_dfs[i]
        cur_score_df = get_score_df(res, in_df, concept, concept_id, text_col, doc_id_col, get_highlights)
        if score_df is None:
            score_df = cur_score_df
        else:
            score_df = pd.concat([score_df, cur_score_df])

    save_progress(sess, score_df, step_name="score_helper", start=None, res=res_full, model_name=model_name)
    if sess is not None:
        # Save to session if provided
        sess.results[concept_id] = score_df
    
        # Generate summary
        cur_summary = await summarize_concept(score_df, concept_id, model_name, sess=sess)
    return score_df

# Performs scoring for all concepts
# Input: concepts, text_df (columns: doc_id, text)
#   --> text could be original, filtered (quotes), and/or summarized (bullets)
# Parameters: threshold
# Output: score_df (columns: doc_id, text, concept_id, concept_name, concept_prompt, score, highlight)
async def score_concepts(text_df, text_col, doc_id_col, concepts, model_name="gpt-3.5-turbo", batch_size=5, get_highlights=False, sess=None):
    # Scoring operates on "text" column for each concept
    start = time.time()

    text_df = text_df.copy()
    # Filter to non-empty rows
    text_df = filter_empty_rows(text_df, text_col)

    text_df[doc_id_col] = text_df[doc_id_col].astype(str)
    tasks = [score_helper(concept, concept_i, concept_id, text_df, text_col, doc_id_col, model_name, batch_size, get_highlights, sess=sess) for concept_i, (concept_id, concept) in enumerate(concepts.items())]
    score_dfs = await tqdm_asyncio.gather(*tasks, file=sys.stdout)

    # Combine score_dfs
    score_df = pd.concat(score_dfs)

    save_progress(sess, score_df, step_name="score", start=start, res=None, model_name=None)
    return score_df

# Based on concept scoring, refine concepts
# Input: score_df (columns: doc_id, text, concept_id, concept_name, concept_prompt, score, highlight), concepts
# Parameters: 
# - threshold (float): minimum score of positive class
# - generic_threshold (float): min fraction concept matches to be considered generic
# - rare_threshold (float): max fraction of concept matches considered rare
# Output: 
# - concepts (dict)
def refine(score_df, concepts, threshold=1, generic_threshold=0.75, rare_threshold=0.05, debug=True):
    # Check for generic, rare, and redundant concepts
    # TODO: add support for redundant concepts
    concepts = concepts.copy()  # Make a copy of the concepts dict to avoid modifying the original
    generic = []
    rare = []
    concept_ids = score_df["concept_id"].unique().tolist()
    for c_id in concept_ids:
        cur_concept_df = score_df[score_df["concept_id"] == c_id]
        cur_concept_pos = score_df[(score_df["concept_id"] == c_id) & (score_df["score"] >= threshold)]
        # Get fraction of positive labels out of all examples
        pos_frac = len(cur_concept_pos) / len(cur_concept_df)
        if pos_frac >= generic_threshold:
            if debug:
                print(f"Generic: {concepts[c_id]['name']}, {pos_frac} match")
            generic.append(c_id)
        elif pos_frac < rare_threshold:
            if debug:
                print(f"Rare: {concepts[c_id]['name']}, {pos_frac} match")
            rare.append(c_id)

    # Remove identified concepts
    if debug:
        print(f"Generic ({len(generic)}): {[concepts[c_id]['name'] for c_id in generic]}")
        print(f"Rare ({len(rare)}): {[concepts[c_id]['name'] for c_id in rare]}")
    for c_id in generic:
        concepts.pop(c_id, None)
    for c_id in rare:
        concepts.pop(c_id, None)
    
    return concepts

async def summarize_concept(score_df, concept_id, model_name="gpt-4-turbo-preview", sess=None, threshold=0.75, summary_length="15-20 word", score_col="score", highlight_col="highlight"):
    # Summarizes behavior in each concept
    df = score_df.copy()
    df = df[df[score_col] >= threshold]

    # Prepare inputs
    arg_dicts = []
    cur_df = df[df["concept_id"] == concept_id]
    cur_df = cur_df.sample(frac=1)  # shuffle order
    if len(cur_df) == 0:
        # No concept matches to summarize
        return None
    concept_name = cur_df["concept_name"].iloc[0]
    concept_prompt = cur_df["concept_prompt"].iloc[0]

    arg_dicts = [{
        "concept_name": concept_name,
        "concept_prompt": concept_prompt,
        "examples": cur_df[highlight_col].tolist(),
        "summary_length": summary_length
    }]
    
    # Run prompts
    prompt_template = summarize_concept_prompt
    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)

    # Process results
    res = res_text[0]
    cur_summary = json_load(res, top_level_key="summary")
    if cur_summary is not None:
        if sess is not None:
            sess.concepts[concept_id].summary = cur_summary
        else:
            return cur_summary
    return None

# Returns ids of not-covered documents
def get_not_covered(score_df, doc_id_col, threshold=0.75, debug=False):
    # Convert to thresholded scores; sum scores across concepts for each doc
    df = score_df.copy()
    df["score"] = df["score"].apply(lambda x: 1 if x >= threshold else 0)
    df = df.groupby(doc_id_col).sum().reset_index() # Sum scores for each doc
    if debug:
        display(df)

    # Filter to examples with no positive scores
    df = df[df["score"] == 0]
    doc_ids = df[doc_id_col].tolist()
    if debug:
        print(doc_ids)
    return doc_ids

# Returns ids of covered-by-generic documents
def get_covered_by_generic(score_df, doc_id_col, threshold=0.75, generic_threshold=0.5, debug=False):
    # Determines generic concepts
    df = score_df.copy()
    df["score"] = df["score"].apply(lambda x: 1 if x >= threshold else 0)
    df_generic = df.groupby("concept_id").mean().reset_index()
    df_generic.rename(columns={"score": "pos_frac"}, inplace=True)
    df_generic = df_generic[df_generic["pos_frac"] >= generic_threshold]
    generic_concepts = df_generic["concept_id"].unique().tolist()
    if debug:
        display(df_generic)
        print(generic_concepts)

    # Determines covered-by-generic documents (those that only match generic concepts)
    # Remove rows for generic concepts and filter to examples with no positive scores
    df_out = score_df.copy()
    df_out = df_out[~df_out["concept_id"].isin(generic_concepts)]
    df_out = df_out.groupby(doc_id_col).sum().reset_index() # Sum scores for each doc
    df_out = df_out[df_out["score"] == 0]
    doc_ids = df_out[doc_id_col].tolist()
    if debug:
        print(doc_ids)
    return doc_ids

# Input: score_df (columns: doc_id, doc, concept_id, score)
# Output: text_df (columns: doc_id, doc) --> filtered set of documents to run further iterations of concept induction
# Returns None if (1) there are zero remaining documents after filtering or (2) there are no changes in the number of documents after filtering
def loop(score_df, doc_col, doc_id_col, debug=False):
    # Check for not-covered and covered-by-generic documents
    # Not-covered: examples that don't match any concepts
    # Covered-by-generic: examples that only match generic concepts (those which cover the majority of examples)
    # TODO: Allow users to decide on custom filtering conditions
    # TODO: Save generic concepts to session (to avoid later)
    n_initial = len(score_df[doc_id_col].unique())

    underrep_ids = get_not_covered(score_df, doc_id_col)
    generic_ids = get_covered_by_generic(score_df, doc_id_col)
    ids_to_include = underrep_ids + generic_ids
    ids_to_include = set(ids_to_include)
    if debug:
        print(f"ids_to_include ({len(ids_to_include)}): {ids_to_include}")

    text_df = score_df.copy()
    text_df = text_df[text_df[doc_id_col].isin(ids_to_include)][[doc_id_col, doc_col]].drop_duplicates().reset_index()

    # Stopping condition: if text_df is empty or has the same number of docs as the original df
    n_final = len(text_df[doc_id_col].unique())
    if (n_final == n_initial) or (len(text_df) == 0):
        return None
    return text_df


def trace():
    # Input: concept_df (columns: doc_id, text, concept_id, ...), text_dfs (columns: doc_id, text)
    # Output: trace_df (columns: doc_id, text, concept_id, score, text1, text2)

    # Joins the score_df with other text dfs that share the same doc_id column
    # To trace from the final concepts and scores to the original text
    pass

def parse_tf_answer(ans):
    if ans.lower() == "true":
        return True
    elif ans.lower() == "false":
        return False
    return False

def clean_item_id(x):
    x = x.replace("item_id ", "")
    x = x.strip()
    return x
    
# Eval helpers
# Input: items (dict), concepts (list of strings)
# Output: concepts_found (dict: concept_id -> list of items), concept_coverage (float)
async def auto_eval(items, concepts, model_name="gpt-3.5-turbo", debug=False, sess=None):
    # Iterate through all concepts to check whether they match any of the items
    start = time.time()

    items_str = "\n".join([f"- item_id {item_id}: {item['name']}. {item['prompt']}" for item_id, item in items.items()])
    items_dict = {str(item_id): item for item_id, item in items.items()}
    item_names_dict = {item['name']: item for _, item in items.items()}
    concepts_dict = {str(i): c for i, c in enumerate(concepts)}
    concepts_str = "\n".join([f"- concept_id {c_id}: {concept}" for c_id, concept in concepts_dict.items()])

    arg_dicts = [
        {
            "concepts": concepts_str,
            "items": items_str,
        }
    ]

    if debug:
        print(arg_dicts)

    # Run prompts
    prompt_template = concept_auto_eval_prompt
    res_text, res_full = await multi_query_gpt_wrapper(prompt_template, arg_dicts, model_name)
    
    res_text = res_text[0]
    if debug:
        print(res_text)

    # Return the concept coverage and mapping from concepts to items
    concepts_found = {}
    concepts_found_unique = {}
    items_added = []
    matches = json_load(res_text, top_level_key="concept_matches")
    if matches is not None:
        for match in matches:
            concept_id = str(match["concept_id"])
            item_id = str(match["item_id"])
            if "item_id" in item_id:
                item_id = clean_item_id(item_id)
            if (concept_id not in concepts_dict) or (item_id not in items_dict):
                continue
            item = items_dict[item_id]
            concepts_found[concept_id] = item
            if item not in items_added:
                concepts_found_unique[concept_id] = item
                items_added.append(item)
    
    if (len(concepts_found) == 0) and (matches is not None):
        # Check for case where items were provided in plain text
        for match in matches:
            concept_id = str(match["concept_id"])
            item_name = str(match["item_id"])
            if (concept_id not in concepts_dict) or (item_name not in item_names_dict):
                continue
            item = item_names_dict[item_name]
            concepts_found[concept_id] = item
            if item not in items_added:
                concepts_found_unique[concept_id] = item
                items_added.append(item)
    
    # Display metrics
    concept_coverage = len(concepts_found_unique) / len(concepts)
    print(f"Concept coverage: {np.round(concept_coverage, 2)}")

    print("Concepts found:")
    for c_id, c in concepts_dict.items():
        if c_id in concepts_found:
            print(f"{c_id}: {c} -- {concepts_found[c_id]['name']}: {concepts_found[c_id]['prompt']}")
        else:
            print(f"{c_id}: {c} -- NONE")

    save_progress(sess, df=None, step_name="auto_eval", start=start, res=res_full, model_name=model_name)
    return concepts_found, concept_coverage

# Adds color-background styling for score columns to match score value
def format_scores(x: float):
    color_str = f"rgba(130, 193, 251, {x*0.5})"
    start_tag = f"<div class='score-col' style='background-color: {color_str}'>"
    return f"{start_tag}{x}</div>"

# Converts a list of bullet strings to a formatted unordered list
def format_bullets(orig, add_quotes=False):
    if (not isinstance(orig, list)) or len(orig) == 0:
        return ""
    if add_quotes:
        lines = [f"<li>\"{line}\"</li>" for line in orig]
    else:
        lines = [f"<li>{line}</li>" for line in orig]
    return "<ul>" + "".join(lines) + "</ul>"

# Adds color-background styling for highlight columns to match score value
def format_highlight(orig: str, quotes, x: float):
        color_str = f"rgba(130, 193, 251, {x*0.5})"
        start_tag = f"<span style='background-color: {color_str}'>"
        end_tag = "</span>"
        if (not isinstance(quotes, str)):
            # Skip if quotes are not a valid string
            return orig
        quotes = quotes.split("\n")
        for quote in quotes:
            quote = str(quote)  # Cast result to string
            orig = str(orig)
            if quote in orig:
                orig = orig.replace(quote, f"{start_tag}{quote}{end_tag}")
        return orig

# Transforms scores to be between 0 and 1
def clean_score(x, threshold):
    if pd.isna(x):
        return 0.0
    elif (x is True) or (x == "True"):
        return 1.0
    elif (x is False) or (x == "False"):
        return 0.0
    else:
        if threshold is None:
            return x
        if x < threshold:
            return 0.0
        return 1.0

# Check if the given example (represented by a row) is an outlier (i.e., doesn't match any of the concepts)
def is_outlier(row, concept_names, threshold):
    concept_matches = [clean_score(row[c], threshold) for c in concept_names]
    return not np.any(concept_matches)

# Helper function for `prep_vis_dfs()` to generate a dataframe with a column containing the scores for each concept
# The output dataframe has: doc_id_col, doc_col, any specified cols_to_show, and a column for each concept (where the column name is the concept name)
def get_concept_col_df(df, score_df, concepts, doc_id_col, doc_col, score_col, cols_to_show):
    id_cols = [doc_id_col, doc_col] + cols_to_show
    cur_df = df[id_cols].copy()
    concept_cols_to_show = [doc_id_col, score_col]
    for c_id, c in concepts.items():
        c_df = score_df[score_df["concept_id"] == c_id][concept_cols_to_show]
        c_df = c_df.rename(columns={score_col: c.name}) # Store score under concept name
        # Rename columns and remove unused columns
        cur_df[doc_id_col] = cur_df[doc_id_col].astype(str)
        c_df[doc_id_col] = c_df[doc_id_col].astype(str)
        cur_df = cur_df.merge(c_df, on=doc_id_col, how="left")
    return cur_df

# Helper function for `visualize()` to generate the underlying dataframes
# Parameters:
# - threshold: float (minimum score of positive class)
def prep_vis_dfs(df, score_df, doc_id_col, doc_col, score_col, df_filtered, df_bullets, concepts, cols_to_show, custom_groups, show_highlights, norm_by, debug=False, threshold=None, outlier_threshold=0.75):
    # TODO: codebook info

    # Handle groupings
    # Add the "All" grouping by default
    groupings = {
        "All": {"x": None, "fn": None},
    }
    if len(custom_groups) > 0:
        # Add custom groupings
        groupings.update(custom_groups)
        for group_name, group_info in custom_groups.items():
            grouping_col = group_info["x"]
            if grouping_col not in cols_to_show:
                cols_to_show.append(grouping_col)

    # Fetch the results table
    df = get_concept_col_df(df, score_df, concepts, doc_id_col, doc_col, score_col, cols_to_show)
    df[doc_id_col] = df[doc_id_col].astype(str)  # Ensure doc_id_col is string type
    # cb = self.get_codebook_info()

    concept_cts = {}
    group_cts = {}
    concept_names = [c.name for c in concepts.values()]
    df["Outlier"] = [is_outlier(row, concept_names, outlier_threshold) for _, row in df.iterrows()]
    concept_names.append("Outlier")
    concept_cts["Outlier"] = sum(df["Outlier"])

    item_metadata = {}
    matrix_df_rows = []
    item_df = None
    item_df_wide = None

    df_bullets[doc_id_col] = df_bullets[doc_id_col].astype(str)
    df_bullets = df_bullets.groupby(doc_id_col).agg(lambda x: list(x)).reset_index()

    # Rationale df
    rationale_col = "rationale"
    highlight_col = "highlight"
    rationale_df = score_df[[doc_id_col, "concept_name", rationale_col, highlight_col]]
    rationale_df[doc_id_col] = rationale_df[doc_id_col].astype(str)

    # Prep data for each group
    for group_name, group_filtering in groupings.items():
        filter_x = group_filtering["x"]
        filter_func = group_filtering["fn"]
        if filter_func is None:
            group_matches = [True] * len(df)
        else:
            group_matches = [filter_func(x) for x in df[filter_x].tolist()]
        cur_df = df[group_matches]
        group_cts[group_name] = len(cur_df)

        def get_text_col_and_rename(orig_df, doc_id_col, new_col_name):
            # Internal helper to get the text column and rename it
            candidate_cols = [c for c in orig_df.columns if c != doc_id_col]
            if len(candidate_cols) != 1:
                raise ValueError(f"Expected 1 text column, got {len(candidate_cols)}")
            orig_col_name = candidate_cols[0]
            orig_df = orig_df.rename(columns={orig_col_name: new_col_name})
            return orig_df

        # Match with filtered example text
        filtered_ex_col = "quotes"
        df_filtered = get_text_col_and_rename(df_filtered, doc_id_col, new_col_name=filtered_ex_col)
        df_filtered[doc_id_col] = df_filtered[doc_id_col].astype(str)
        cur_df = cur_df.merge(df_filtered, on=doc_id_col, how="left")

        # Match with bullets
        bullets_col = "bullets"
        df_bullets = get_text_col_and_rename(df_bullets, doc_id_col, new_col_name=bullets_col)
        cur_df = cur_df.merge(df_bullets, on=doc_id_col, how="left")

        # Item df
        item_df_cols = [doc_id_col, doc_col, filtered_ex_col, bullets_col] + cols_to_show
        cur_item_df = pd.melt(cur_df, id_vars=item_df_cols, value_vars=concept_names, var_name="concept", value_name="concept score")
        cur_item_df["id"] = group_name # cols: ex_col, concept, concept_score, id (cluster id)
        cur_item_df["concept_score_orig"] = [clean_score(x, threshold) for x in cur_item_df["concept score"].tolist()]
        cur_item_df["concept score"] = [format_scores(x) for x in cur_item_df["concept_score_orig"].tolist()]

        # Format bullets
        cur_item_df[bullets_col] = [format_bullets(orig) for orig in cur_item_df[bullets_col]]

        # Add rationale
        cur_item_df = cur_item_df.merge(rationale_df, left_on=[doc_id_col, "concept"], right_on=[doc_id_col, "concept_name"], how="left")

        # Add highlight styling
        if show_highlights:
            cur_item_df[doc_col] = cur_item_df.apply(lambda x: format_highlight(x[doc_col], x[highlight_col], x["concept_score_orig"]), axis=1)

        # Format text
        cur_df[doc_col] = [textwrap.fill(orig, width=50) for orig in cur_df[doc_col]]

        if debug:
            cur_item_df = cur_item_df[["id", "concept", "concept score", "concept_score_orig", doc_col, filtered_ex_col, bullets_col, rationale_col] + cols_to_show]
        else:
            cur_item_df = cur_item_df[["id", "concept", "concept score", "concept_score_orig", doc_col, bullets_col, rationale_col] + cols_to_show]
        if item_df is None:
            item_df = cur_item_df
        else:
            item_df = pd.concat([item_df, cur_item_df])
        
        # Item wide df
        cur_item_df_wide = cur_df.copy()
        cur_item_df_wide["id"] = group_name
        for concept in concept_names:
            cur_item_df_wide[concept] = [clean_score(x, threshold) for x in cur_item_df_wide[concept].tolist()]
            cur_item_df_wide[concept] = [format_scores(x) for x in cur_item_df_wide[concept].tolist()]

        cols_to_include = ["id", doc_col] + cols_to_show + concept_names
        cur_item_df_wide = cur_item_df_wide[cols_to_include]
        if item_df_wide is None:
            item_df_wide = cur_item_df_wide
        else:
            item_df_wide = pd.concat([item_df_wide, cur_item_df_wide])

        # Metadata
        cluster_avg_overall_score = NAN_SCORE  # TEMPLATE SCORE for datasets without scores
        item_metadata[group_name] = {
            "Slice size": f"{len(cur_df)} examples",
        }

        # Matrix df
        for concept in concept_names:
            cur_scores = [clean_score(x, threshold) for x in cur_df[concept].tolist()]
            if len(cur_scores) > 0:
                matches = [x for x in cur_scores if x > outlier_threshold]
                n_matches = len(matches)
            else:
                n_matches = 0
            
            if group_name == "All":
                concept_cts[concept] = n_matches

            matrix_row = [group_name, n_matches, group_name, cluster_avg_overall_score, concept]
            matrix_df_rows.append(matrix_row)

    matrix_df = pd.DataFrame(matrix_df_rows, columns=["id", "value", "example", "_my_score", "concept"])

    # Perform normalization
    def calc_norm_by_group(row):
        g = row["id"]
        group_ct = group_cts[g]
        if group_ct == 0:
            return 0
        return row["value"] / group_ct
        
    def calc_norm_by_concept(row, val_col):
        c = row["concept"]
        concept_ct = concept_cts[c]
        if concept_ct == 0:
            return 0
        return row[val_col] / concept_ct

    if norm_by == "group":
        # Normalize by group
        matrix_df["value"] = [calc_norm_by_group(row) for _, row in matrix_df.iterrows()]
    else:
        # Normalize by concept
        matrix_df["value"] = [calc_norm_by_concept(row, "value") for _, row in matrix_df.iterrows()]
    
    # Replace any 0s with NAN_SCORE
    matrix_df["value"] = [NAN_SCORE if x == 0 else x for x in matrix_df["value"].tolist()]

    # Metadata
    def get_concept_metadata(c):
        ex_ids = c.example_ids
        
        if len(ex_ids) > 0:
            cur_df = score_df[(score_df["concept_id"] == c.id) & (score_df[doc_id_col].isin(ex_ids))]
            ex = cur_df[highlight_col].unique().tolist()
        else:
            ex = []
        res = {
            "Criteria": f"<br>{c.prompt}",
            "Summary": f"<br>{c.summary}",
            "Matches": f"{concept_cts[c.name]} examples",
            "Representative examples": f"{format_bullets(ex, add_quotes=True)}"
        }
        return res
    
    concept_metadata = {c.name: get_concept_metadata(c) for c in concepts.values()}
    concept_metadata["Outlier"] = {
        "Criteria": OUTLIER_CRITERIA,
        "Concept matches": f"{concept_cts['Outlier']} examples",
    }
    metadata_dict = {
        "items": item_metadata,
        "concepts": concept_metadata,
    }

    return matrix_df, item_df, item_df_wide, metadata_dict

# Generates the in-notebook visualization for concept induction results
#
# Input:
# - in_df: DataFrame (includes columns: doc_id_col, doc_col)
# - score_df: DataFrame (columns: doc_id_col, text, concept_id, concept_name, concept_prompt, score, highlight)
# - doc_col: string (column name for full document text)
# - doc_id_col: string (column name for document ID)
# - score_col: string (column name for concept score)
# - df_filtered: DataFrame (columns: doc_id_col, text_col)
# - df_bullets: DataFrame (columns: doc_id_col, text_col)
# - concepts: dict (concept_id -> Concept object)
#
# Parameters:
# - cols_to_show: list of strings (column names to show in visualization)
# - custom_groups: dict (group_name -> group information)
# - show_highlights: boolean (whether to show highlights)
# - norm_by: string (column name to normalize by; either "group" or "concept")
# - debug: boolean (whether to print debug statements)
def visualize(in_df, score_df, doc_col, doc_id_col, score_col, df_filtered, df_bullets, concepts, cols_to_show=[], custom_groups={}, show_highlights=False, norm_by="concept", debug=False):
    matrix_df, item_df, item_df_wide, metadata_dict = prep_vis_dfs(in_df, score_df, doc_id_col, doc_col, score_col, df_filtered, df_bullets, concepts, cols_to_show=cols_to_show, custom_groups=custom_groups, show_highlights=show_highlights, norm_by=norm_by, debug=debug)

    data = matrix_df.to_json(orient='records')
    data_items = item_df.to_json(orient='records')
    data_items_wide = item_df_wide.to_json(orient='records')
    md = json.dumps(metadata_dict)
    w = MatrixWidget(
        data=data, 
        data_items=data_items,
        data_items_wide=data_items_wide, 
        metadata=md
    )
    return w, matrix_df, item_df, item_df_wide

def get_select_widget(concepts_json):
    w = ConceptSelectWidget(
        data=concepts_json,
    )
    return w

# Edits an existing concept with the specified ID
# Input: 
# - concepts: dict (concept_id -> concept dict) with existing set of concepts
# - concept_id: string (ID of the concept to edit)
# Parameters:
# - new_name: string (Optional, new name of the concept)
# - new_prompt: string (Optional, new prompt for the concept)
# - new_ex_ids: list of strings (Optional, IDs of examples that match the new concept)
# Output:
# - concepts: dict (concept_id -> concept dict) with the concept edited
def edit_concept(concepts, concept_id, new_name=None, new_prompt=None, new_ex_ids=None):
    # Update dictionary
    cur_concept = concepts[concept_id]
    if new_name is not None:
        cur_concept["name"] = new_name
    if new_prompt is not None:
        cur_concept["prompt"] = new_prompt
    if new_ex_ids is not None:
        cur_concept["example_ids"] = new_ex_ids
        # TODO: differentiate between replacing and appending example IDs
    concepts[concept_id] = cur_concept

    # TODO: handle concept_df
    return concepts



# Distill - Filter ========================
filter_prompt = """
I have the following TEXT EXAMPLE:
{ex}

Please extract {n_quotes} QUOTES exactly copied from this EXAMPLE that are {seeding_phrase}. Please respond ONLY with a valid JSON in the following format:
{{
    "relevant_quotes": [ "<QUOTE_1>", "<QUOTE_2>", ... ]
}}
"""
# Removed: If there are no quotes relevant to {seed}, leave the list empty.

# Distill - Summarize ========================
summarize_prompt = """
I have the following TEXT EXAMPLE:
{ex}

Please summarize the main point of this EXAMPLE {seeding_phrase} into {n_bullets} bullet points, where each bullet point is a {n_words} word phrase. Please respond ONLY with a valid JSON in the following format:
{{
    "bullets": [ "<BULLET_1>", "<BULLET_2>", ... ]
}}
"""

# Synthesize ========================
synthesize_prompt = """
I have this set of bullet point summaries of text examples:
{examples}

Please write a summary of {n_concepts_phrase} for these examples. {seed_phrase} For each high-level pattern, write a 2-4 word NAME for the pattern and an associated 1-sentence ChatGPT PROMPT that could take in a new text example and determine whether the relevant pattern applies. Also include 1-2 example_ids for items that BEST exemplify the pattern. Please respond ONLY with a valid JSON in the following format:
{{
    "patterns": [ 
        {{"name": "<PATTERN_NAME_1>", "prompt": "<PATTERN_PROMPT_1>", "example_ids": ["<EXAMPLE_ID_1>", "<EXAMPLE_ID_2>"]}},
        {{"name": "<PATTERN_NAME_2>", "prompt": "<PATTERN_PROMPT_2>", "example_ids": ["<EXAMPLE_ID_1>", "<EXAMPLE_ID_2>"]}},
    ]
}}
"""

# Review ========================
review_remove_prompt = """
I have this set of themes generated from text examples:
{themes}

Please identify any themes that should be REMOVED because they are either:
(1) Too specific/narrow and would only describe a few examples, or 
(2) Too generic/broad and would describe nearly all examples.
If there no such themes, please leave the list empty.
Please respond ONLY with a valid JSON in the following format:

{{
    "remove": [ 
        "<THEME_NAME_5>",
        "<THEME_NAME_6>",
    ]
}}
"""

review_merge_prompt = """
I have this set of themes generated from text examples:
{themes}

Please identify any PAIRS of themes that are similar or overlapping that should be MERGED together. 
Please respond ONLY with a valid JSON in the following format with the original themes and a new name and prompt for the merged theme. Do NOT simply combine the prior theme names or prompts, but come up with a new 2-3 word name and 1-sentence ChatGPT prompt. If there no similar themes, please leave the list empty.

{{
    "merge": [ 
        {{
            "original_themes": ["<THEME_NAME_A>", "<THEME_NAME_B>"],
            "merged_theme_name": "<THEME_NAME_AB>",
            "merged_theme_prompt": "<THEME_PROMPT_AB>",
        }},
        {{
            "original_themes": ["<THEME_NAME_C>", "<THEME_NAME_D>"],
            "merged_theme_name": "<THEME_NAME_CD>",
            "merged_theme_prompt": "<THEME_PROMPT_CD>",
        }}
    ]
}}
"""

# Score ========================
score_no_highlight_prompt = """
CONTEXT: 
    I have the following text examples in a JSON:
    {examples_json}

    I also have a pattern named {pattern_name} with the following PROMPT: 
    {pattern_prompt}

TASK:
    For each example, please evaluate the PROMPT by generating a 1-sentence RATIONALE of your thought process and providing a resulting ANSWER of ONE of the following multiple-choice options, including just the letter: 
    - A: Strongly agree
    - B: Agree
    - C: Neither agree nor disagree
    - D: Disagree
    - E: Strongly disagree
    Respond with ONLY a JSON with the following format, escaping any quotes within strings with a backslash:
    {{
        "pattern_results": [
            {{
                "example_id": "<example_id>",
                "rationale": "<rationale>",
                "answer": "<answer>",
            }}
        ]
    }}
"""

score_highlight_prompt = """
CONTEXT: 
    I have the following text examples in a JSON:
    {examples_json}

    I also have a pattern named {pattern_name} with the following PROMPT: 
    {pattern_prompt}

TASK:
    For each example, please evaluate the PROMPT by generating a 1-sentence RATIONALE of your thought process and providing a resulting ANSWER of ONE of the following multiple-choice options, including just the letter: 
    - A: Strongly agree
    - B: Agree
    - C: Neither agree nor disagree
    - D: Disagree
    - E: Strongly disagree
    Please also include one 1-sentence QUOTE exactly copied from the example that illustrates this pattern.
    Respond with ONLY a JSON with the following format, escaping any quotes within strings with a backslash:
    {{
        "pattern_results": [
            {{
                "example_id": "<example_id>",
                "rationale": "<rationale>",
                "answer": "<answer>",
                "quote": "<quote>"
            }}
        ]
    }}
"""

score_overall_topic_prompt = """
CONTEXT: 
    I have the following text examples in a JSON:
    {examples_json}

    I also have a pattern named {pattern_name} with the following PROMPT: 
    AS ITS PRIMARY TOPIC, {pattern_prompt}

TASK:
    For each example, please evaluate the PROMPT by generating a 1-sentence RATIONALE of your thought process and providing a resulting ANSWER of ONE of the following multiple-choice options, including just the letter: 
    - A: Strongly agree
    - B: Agree
    - C: Neither agree nor disagree
    - D: Disagree
    - E: Strongly disagree
    Only answer with "A" if the example is PRIMARILY about the topic.
    Respond with ONLY a JSON with the following format, escaping any quotes within strings with a backslash:
    {{
        "pattern_results": [
            {{
                "example_id": "<example_id>",
                "rationale": "<rationale>",
                "answer": "<answer>",
            }}
        ]
    }}
"""

# Summarize Concept ========================
summarize_concept_prompt = """
Please write a BRIEF {summary_length} executive summary of the theme "{concept_name}" as it appears in the following examples.
{examples}

DO NOT write the summary as a third party using terms like "the text examples" or "they discuss", but write the summary from the perspective of the text authors making the points directly.
Please respond ONLY with a valid JSON in the following format:
{{
    "summary": "<SUMMARY>"
}}
"""

concept_auto_eval_prompt = """
I have this set of CONCEPTS:
{concepts}

I have this set of TEXTS: 
{items}

Please match at most ONE TEXT to each CONCEPT. To perform a match, the text must EXACTLY match the meaning of the concept. Do NOT match the same TEXT to multiple CONCEPTS.

Here are examples of VALID matches:
- Global Diplomacy, International Relations; rationale: "The text is about diplomacy between countries."
- Statistical Data, Quantitative Evidence; rationale: "The text is about data and quantitative measures."
- Policy and Regulation, Policy issues and legislation; rationale: "The text is about policy, laws, and legislation."

Here are examples of INVALID matches:
- Reputation Impact, Immigration
- Environment, Politics and Law
- Interdisciplinary Politics, Economy

If there are no valid matches, please EXCLUDE the concept from the list. Please provide a 1-sentence RATIONALE for your decision for any matches. 
Please respond with a list of each concept and either the item it matches or NONE if no item matches in this format:
{{
    "concept_matches": [
        {{
            "concept_id": "<concept_id_number>",
            "item_id": "<item_id_number or NONE>",
            "rationale": "<rationale for match>",
        }}
    ]
}}
"""
